@@ -10,7 +10,6 @@
 #endif
 
 #define WITH_RESOLV
-#undef DO_BIND
 
 /*
  * Control glibc, especially to get 64 bit file offsets
@@ -140,7 +139,7 @@
 #include <netdb.h>
 #endif
 
-#if !defined(HAVE_SOCKLEN_T) && !defined(CRAENV) && !defined(NORESOLV)
+#if !defined(HAVE_SOCKLEN_T) && !defined(NORESOLV)
 typedef u_int32_t socklen_t;
 #endif
 
@@ -186,6 +185,7 @@ struct options {
   int S;
   int T;
   int user_specified_blocksize;
+  int six;
 };
 
 void malloc_page_aligned(const struct options *o, struct pmalloc * pmalloc);
@@ -209,24 +209,28 @@ struct progstate {
   int teefd; /* Usually -1, but can be a file descriptor number to copy
 	      * the stream to.  Currently that will always be 1 (stdout) */
   int using_o_direct;
+  int using_o_direct_i; /* Input */
   struct pmalloc pmalloc; /* Size is b (blocksize) */
 };
 
 #ifndef NOTIMER 
-static void sigtimer(const int signal)
+static void
+sigtimer(const int signal)
 {
   signal_report = 1;
 }
 #endif
 
-static void sigshutdown(const int signal)
+static void
+sigshutdown(const int signal)
 {
   stopit = 1;
 }
 
 #if 0
 /* Disable that for now */
-static void sigchld(const int sig)
+static void
+sigchld(const int sig)
 {
   int status;
   int ret;
@@ -264,19 +268,22 @@ static void sigchld(const int sig)
 }
 #endif
 
-static void sigreport(int signal)
+static void
+sigreport(int signal)
 {
   signal_report = 1;
 }
 
-static void sighup(int signal)
+static void
+sighup(int signal)
 {
-  write(2, "SIGHUP\n", sizeof("SIGHUP\n")-1);
+  int meh = write(2, "SIGHUP\n", sizeof("SIGHUP\n") - 1);
   n_sighups++;
 }
 
 #ifdef SIGHUP
-static void handle_sighup(void)
+static void
+handle_sighup(void)
 {
   fprintf(stderr,
 	  "Received SIGHUP. "
@@ -294,14 +301,16 @@ static void handle_sighup(void)
 }
 #endif
 
-static struct options *default_options(struct options *const o)
+static struct options *
+default_options(struct options *const o)
 {
   bzero(o, sizeof(struct options));
   return o;
 }
 
 #ifdef HAVE_STRUCT_SIGACTION
-static void tsignal(const int sig, void (*const handler)(int))
+static void
+tsignal(const int sig, void (*const handler)(int))
 {
   struct sigaction sa;
 
@@ -314,7 +323,8 @@ static void tsignal(const int sig, void (*const handler)(int))
   }
 }
 #else
-static void tsignal(const int sig, void (*const handler)(int))
+static void
+tsignal(const int sig, void (*const handler)(int))
 {
   if (signal(sig, handler) == SIG_ERR) {
     perror("signal");
@@ -325,7 +335,8 @@ static void tsignal(const int sig, void (*const handler)(int))
 
 #ifdef WANT_SOUND
 
-static void setaudio(int fd, const char *spec)
+static void
+setaudio(int fd, const char *spec)
 {
   struct soundoptions {
     int so_format;
@@ -361,7 +372,8 @@ static void setaudio(int fd, const char *spec)
 	    , spec, strerror(errno));
 }
 #else /* WANT_SOUND */
-static void setaudio(int fd, const char *const spec)
+static void
+setaudio(int fd, const char *const spec)
 {
   fprintf(stderr, "Sorry, don't have soundcard support compiled in, "
 	  "no audio settings will be made.\nTrying to play anyway.\n");
@@ -369,7 +381,8 @@ static void setaudio(int fd, const char *const spec)
 #endif /* WANT_SOUND */
 
 #ifdef NOTCP
-static int open_tcp(const struct options *const o, int mode)
+static int
+open_tcp(const struct options *const o, int mode)
 {
   fprintf(stderr, "Sorry, TCP/IP socket support has not been compiled in.\n");
   exit(1);
@@ -377,7 +390,8 @@ static int open_tcp(const struct options *const o, int mode)
 
 #else /* NOTCP */
 
-static void print_inet(FILE *const f, const void *const raw)
+static void
+print_inet(FILE *const f, const void *const raw)
 {
   const unsigned char *c = raw;
 
@@ -385,24 +399,18 @@ static void print_inet(FILE *const f, const void *const raw)
 	  , (int)*(c + 2), (int)*(c + 3));
 }
 
+// this construct's purpose is to always compile the old API
+#ifdef HAVE_GETADDRINFO
+static int use_getaddrlen = 1;
+#else
+static int use_getaddrlen = 0;
+#endif
 
-static int open_tcp(const struct options *const o, int mode)
+int gimme_a_socket(int domain, int type, int protocol)
 {
-  char *hostname;
-  const char *port;
-  int fd = -1;
-  struct hostent *hostent;
-  const char *spec;
-  int newfd;
-  struct sockaddr_in serv_addr;
-  int *size;
+  int fd;
 
-  if (mode == O_WRONLY) {
-    spec = o->o;
-  } else {
-    spec = o->i;
-  }
-  if ((fd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
+  if ((fd = socket(domain, type, protocol)) == -1) {
     perror("socket");
     exit(2);
   }
@@ -413,35 +421,75 @@ static int open_tcp(const struct options *const o, int mode)
       perror("setsockopt(REUSEADDR), [continuing]");
     }
   }
+  return fd;
+}
 
-#ifdef DO_BIND
-  bzero(&serv_addr, sizeof(serv_addr));
-  serv_addr.sin_family = AF_INET;
-  serv_addr.sin_addr.s_addr = inet_addr("192.168.175.73");
-  serv_addr.sin_port = htons(3337);
-  if (bind(fd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) == -1) {
-      perror("bind");
-      exit(2);
+static int
+open_tcp(const struct options *const o, int mode)
+{
+  char *hostname;
+  const char *port;
+  const char *port_iterator;
+  const char *spec;
+  int fd = -1;
+  int newfd = -1;
+  struct sockaddr *use_this_serv_addr = NULL;
+  socklen_t use_this_serv_addr_len = -1;
+
+  if (mode == O_WRONLY) {
+    spec = o->o;
+  } else {
+    spec = o->i;
   }
-#endif
 
-  port = strchr(spec, ':');
+  port = NULL;
+  for (port_iterator = strchr(spec, ':');
+       port_iterator && *port_iterator;
+       port_iterator = strchr(port_iterator, ':')) {
+    port = port_iterator;
+    port_iterator++;
+  }
   port++;
   if (port == NULL) {
     fprintf(stderr, "Can't find port in IP Spec '%s'\n", spec);
     exit(1);
   }
-  if (spec[0] == ':') { /* Listen */
-    /* FIXME, pull in front, so that same code is useable for client */
-    bzero(&serv_addr, sizeof(serv_addr));
-    serv_addr.sin_family = AF_INET;
-    serv_addr.sin_addr.s_addr = INADDR_ANY;
-    serv_addr.sin_port = htons(atoi(port));
-    if (bind(fd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) ==
-	-1) {
+  // this test is to allow connecting to (as a client) '::1:3333'
+  // aka ipv6 localhost
+  if (spec[1] != '\0' && spec[0] == ':' && spec[1] != ':') {
+     /* Listen mode */
+    if (o->six == -1) {
+      /* fixme, pull in front, so that same code is useable for client */
+      struct sockaddr_in serv_addr;
+      use_this_serv_addr = (struct sockaddr *)&serv_addr;
+      use_this_serv_addr_len = sizeof(serv_addr);
+      bzero(&serv_addr, sizeof(serv_addr));
+      serv_addr.sin_family = AF_INET;
+      serv_addr.sin_addr.s_addr = INADDR_ANY;
+      serv_addr.sin_port = htons(atoi(port));
+
+      fd = gimme_a_socket(AF_INET, SOCK_STREAM, 0);
+    } else { // allow or force ipv6
+      struct sockaddr_in6 sin6;
+      bzero(&sin6, sizeof(sin6));
+      use_this_serv_addr = (struct sockaddr *)&sin6;
+      use_this_serv_addr_len = sizeof(sin6);
+      // sin6.sin6_len = sizeof(sin6);
+      sin6.sin6_family = AF_INET6;
+      sin6.sin6_flowinfo = 0;
+      sin6.sin6_port = htons(atoi(port));
+      sin6.sin6_addr = in6addr_any;
+
+      fd = gimme_a_socket(AF_INET6, SOCK_STREAM, 0);
+    }
+    if (o->v > 1 && o->six != -1) 
+      fprintf(stderr, "Bind in IPV6 mode\n");
+    if (bind(fd, (struct sockaddr *)use_this_serv_addr, 
+             use_this_serv_addr_len) == -1) {
       perror("bind");
       exit(2);
     }
+
     if (listen(fd, 2) == -1) {
       perror("listen");
       exit(2);
@@ -450,19 +498,14 @@ static int open_tcp(const struct options *const o, int mode)
     if (o->v >= 1)
       fprintf(stderr, "Accepting on port %d\n", atoi(port));
 
-    bzero(&serv_addr, sizeof(serv_addr));
+    struct sockaddr addr_here;
+    socklen_t addr_here_size;
 
-    size = malloc(sizeof(*size)); /* Yes I know I can stack-allocate instead. */
-    if (size == NULL) {           /* However, a platform refuses accepting the */
-      perror("malloc");           /* size parameter to accept(2) when it is on the */
-      exit(2);                    /* stack and returns EFAULT. */
-    }
-    if ((newfd = accept(fd, (struct sockaddr *)(&serv_addr), size)) == -1) {
+    if ((newfd = accept(fd, &addr_here, &addr_here_size)) == -1) {
       fprintf(stderr, "errno: %d\n", errno);
       perror("accept");
       exit(2);
     }
-    free(size);
     close(fd);
     fd = newfd;
   } else { /* Connect */
@@ -473,48 +516,108 @@ static int open_tcp(const struct options *const o, int mode)
     if (o->v >= 2)
       fprintf(stderr, "Connecting to %s %s\n", hostname, port);
 
+    struct sockaddr_in serv_addr;
+    use_this_serv_addr = (struct sockaddr *)&serv_addr;
+    use_this_serv_addr_len = sizeof(serv_addr);
     bzero(&serv_addr, sizeof(serv_addr));
     serv_addr.sin_family = AF_INET;
     serv_addr.sin_port = htons(atoi(port));
     serv_addr.sin_addr.s_addr = inet_addr(hostname);
-    if (serv_addr.sin_addr.s_addr == INADDR_NONE) {
+    if (serv_addr.sin_addr.s_addr != INADDR_NONE) {
+      // ipv4 address specified
+      fd = gimme_a_socket(AF_INET, SOCK_STREAM, 0);
+    } else {
 #ifdef NORESOLV
       fprintf(stderr, "Not an IP Address and no resolver compiled in: %s\n"
 	      ,hostname);
       exit(1);
 #else
-      struct in_addr **a;
       if (o->v >= 2)
-	fprintf(stderr, "Hostname lookup for '%s'\n", hostname);
-      hostent = gethostbyname(hostname);
-      if (hostent == NULL) {
-	herror(hostname);
-	exit(2);
+	fprintf(stderr, "Hostname lookup for '%s' use getaddrlen? %d\n", 
+                hostname, use_getaddrlen);
+      if (use_getaddrlen == 0) {
+        struct in_addr **a;
+        struct hostent *hostent;
+        hostent = gethostbyname(hostname);
+        if (hostent == NULL) {
+          herror(hostname);
+          exit(2);
+        }
+        a = (struct in_addr **)hostent->h_addr_list;
+        bcopy(*a, &serv_addr.sin_addr.s_addr, sizeof(struct in_addr));
+        if (o->v >= 2) {
+          fprintf(stderr, "Cannonical name %s, type %s, prim addr: "
+                  , hostent->h_name
+                  , hostent->h_addrtype == AF_INET ? "ipv4" : "!ipv4");
+          print_inet(stderr, &serv_addr.sin_addr.s_addr);
+          fprintf(stderr, "\n");
+        }
+        fd = gimme_a_socket(AF_INET, SOCK_STREAM, 0);
       }
-      a = (struct in_addr **)hostent->h_addr_list;
-      bcopy(*a, &serv_addr.sin_addr.s_addr, sizeof(struct in_addr));
-      if (o->v >= 2) {
-	fprintf(stderr, "Cannonical name %s, type %s, prim addr: "
-		, hostent->h_name
-		, hostent->h_addrtype == AF_INET? "ipv4": "!ipv4");
-	print_inet(stderr, &serv_addr.sin_addr.s_addr);
-	fprintf(stderr, "\n");
-#if 0
-	for (; *a; a++) {
-	  print_inet(stderr, *a);
-	  fprintf(stderr, "\n");
-	}
-#endif
+#ifdef HAVE_GETADDRINFO
+      struct addrinfo *result_begin, *result, hints;
+      int error;
+
+      bzero(&hints, sizeof(hints));
+      switch (o->six) {
+      case -1: hints.ai_family = AF_INET; break;
+      case 2: hints.ai_family = AF_INET6; break;
+      default: // usually 1
+        hints.ai_family = PF_UNSPEC; break;
       }
-#endif
-    }
-    if (connect(fd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) ==
-	-1) {
+      hints.ai_socktype = SOCK_STREAM;
+
+      error = getaddrinfo(hostname, port, &hints, &result_begin);
+      if (error) {
+        fprintf(stderr, "getaddrinfo: '%s'\n", gai_strerror(error));
+        exit(3);
+      }
+      for (result = result_begin; result; result = result->ai_next) {
+        if (result->ai_addr)
+          break;
+      }
+      if (result == NULL) {
+        fprintf(stderr, "empty getaddrinfo() result for '%s'\n", hostname);
+        exit(3);
+      }
+      use_this_serv_addr = result->ai_addr;
+      use_this_serv_addr_len = result->ai_addrlen;
+
+      if (o->v > 1) {
+        char *msg;
+        switch (result->ai_family) {
+        case AF_INET: msg = "ipv4"; break;
+        case AF_INET6: msg = "ipv6"; break;
+        default: msg = "not-IPv4/6"; break;
+        }
+        char name[8192];
+        char addr_string[INET6_ADDRSTRLEN];
+        // .... // FIXME - use inet_ntop
+        getnameinfo(use_this_serv_addr, use_this_serv_addr_len,
+                    name, sizeof(name), NULL, 0, 0);
+        if (inet_ntop(result->ai_family, &use_this_serv_addr, addr_string, 
+                      use_this_serv_addr_len) == NULL) {
+          strcpy(addr_string, "unknown");
+        }
+        fprintf(stderr, "Cannonical name '%s', type %s, prim addr: '%s'\n"
+                , name
+                , msg
+                , addr_string);
+      }
+
+      fd = gimme_a_socket(result->ai_family, result->ai_socktype, 
+                          result->ai_protocol);
+
+#endif // HAVE_GETADDRINFO
+    }             
+    if (connect(fd, use_this_serv_addr, use_this_serv_addr_len) == -1) {
       perror("connect");
       exit(2);
     }
     free(hostname);
-  }
+    // FIXME - free getaddrinfo result
+#endif // host name lookup
+}
 
 #if defined(HAVE_GETNAMEINFO) && !defined(NORESOLV)
   if (o->v >= 2) {
@@ -551,6 +654,26 @@ static int open_tcp(const struct options *const o, int mode)
 }
 #endif /* else NOTCP */
 
+int
+get_fs_blocksize(const char *const filename, int flags, int mode)
+{
+      struct statvfs statfs;
+      int fd;
+
+      fd = open(filename, flags, mode);
+      if (fd == -1) {
+	perror("Cannot open input file (to get block size)");
+	exit(2);
+      }
+      if (fstatvfs(fd, &statfs) == -1) {
+	perror("Cannot get FS blocksize, you need to use -b<blocksize>\n");
+	exit(2);
+      }
+      close(fd);
+      return statfs.f_bsize;
+}
+
+// fixme, make this usable as open_input_file, too
 void open_output_file(const struct options *const o
 		      , struct progstate *const state, int flags)
 {
@@ -568,6 +691,7 @@ void open_output_file(const struct options *const o
 #endif
   }
 
+#ifdef HAVE_O_DIRECT
   if (state->using_o_direct) {
     flags |= O_DIRECT;
     if (o->v > 1) {
@@ -577,6 +701,19 @@ void open_output_file(const struct options *const o
 
     if (!o->user_specified_blocksize) {
 #ifdef HAVE_SYS_STATVFS_H
+#if 0
+      // code that broke -B fixme
+      tmp_blocksize = get_fs_blocksize(o->o, flags, mode);
+      if (state->b && state->b != tmp_blocksize) {
+	fprintf(stderr, "WARNING: blocksize set to %d, but output filesystem \n"
+		"requires %d.  Continuing anyway with %d.\n"
+		, state->b
+		, tmp_blocksize
+		, state->b);
+      } else {
+	state->b = tmp_blocksize;
+      }
+#else
       struct statvfs statfs;
 
       state->ofd = open(o->o, flags, mode);
@@ -590,14 +727,16 @@ void open_output_file(const struct options *const o
       }
       state->b = statfs.f_bsize;
       close(state->ofd);
+#endif // broken code
 #else
       fprintf(stderr, "This platform does not have statvfs, you need to "
 	      "specify the blocksize manually with -b, it must match the "
-	      "filesystemblocksize\n");
+	      "output filesystem blocksize\n");
       exit(2);
 #endif
     }
   }
+#endif
 
   state->ofd = open(o->o, flags, mode);
   if (state->ofd == -1) {
@@ -606,8 +745,9 @@ void open_output_file(const struct options *const o
   }
 }
 
-static void init(struct options *const o, struct progstate *const state
-		 , const int blocksize)
+static void
+init(struct options *const o, struct progstate *const state
+     , const int blocksize)
 {
   struct timeval t;
 
@@ -620,8 +760,14 @@ static void init(struct options *const o, struct progstate *const state
   if (o->O == NULL)
     o->O = "";
 
-  if (strchr(o->O, 'D'))
+  if (strchr(o->O, 'D')) {
+#ifdef HAVE_O_DIRECT
     state->using_o_direct = 1;
+#else
+    fprintf(stderr, "Support for O_DIRECT not compiled in\n");
+    exit(1);
+#endif
+  }
 
   if (blocksize == 0) {
     state->b = 8192;
@@ -633,6 +779,12 @@ static void init(struct options *const o, struct progstate *const state
   if (state->b == 0)
     state->b = state->b;
 
+  if (strchr(o->I, 'D'))
+    state->using_o_direct_i = 1;
+  // fixme
+  // need code to check that blocksize specified
+  // works for both in and out
+
   if (o->w == 0)
     o->w = state->b;
 
@@ -667,7 +819,8 @@ static void init(struct options *const o, struct progstate *const state
 	state->ifd = open(o->i, O_RDONLY);
     }
     if (state->ifd == -1) {
-      perror("Cannot open input file");
+      fprintf(stderr, "Cannot open input file/tcpspec '%s': ", o->i);
+      perror(NULL);
       exit(2);
     }
   }
@@ -779,7 +932,8 @@ static void init(struct options *const o, struct progstate *const state
   malloc_page_aligned(o, &state->pmalloc);
 }
 
-static void closefiles(struct progstate *const state)
+static void
+closefiles(struct progstate *const state)
 {
   if (state->ofd != 1 && state->ofd != 2 && state->ofd != -1)
     if (close(state->ofd) == -1) {
@@ -792,9 +946,10 @@ static void closefiles(struct progstate *const state)
 }
 
 /* Will return true if something has been written */
-static void print_kmg(const char *const pre
-		      , const char *const format, const double num
-		      , const char *const post, FILE *const f)
+static void
+print_kmg(const char *const pre
+	  , const char *const format, const double num
+	  , const char *const post, FILE *const f)
 {
   if (pre != NULL)
     fprintf(f, "%s", pre);
@@ -815,9 +970,10 @@ static void print_kmg(const char *const pre
     fprintf(f, "%s", post);
 }
 
-static void report(const struct options *const o,
-		   struct progstate *const state,
-		   int curbytes)
+static void
+report(const struct options *const o,
+       struct progstate *const state,
+       int curbytes)
 {
   struct timeval t2;
   double sofar;
@@ -891,8 +1047,9 @@ static void report(const struct options *const o,
   }
 }
 
-int my_write(const struct options *const o, struct progstate *const state
-	     , const void *buf, const size_t n_bytes) {
+int
+my_write(const struct options *const o, struct progstate *const state
+	 , const void *buf, const size_t n_bytes) {
   int ret;
   static int warning_printed = 0;
 
@@ -948,8 +1105,9 @@ int my_write(const struct options *const o, struct progstate *const state
   exit(2);
 }
 
-void malloc_page_aligned(const struct options *const o
-			 , struct pmalloc *const pmalloc)
+void
+malloc_page_aligned(const struct options *const o
+		    , struct pmalloc *const pmalloc)
 {
   pmalloc->allocated = malloc(pmalloc->n_bytes + pagesize);
 
@@ -969,8 +1127,9 @@ void malloc_page_aligned(const struct options *const o
 }
 
 #if HAVE_POLL_H
-static void pollloop(const struct options *const o
-		     , struct progstate *const state)
+static void
+pollloop(const struct options *const o
+	 , struct progstate *const state)
 {
   char *buf; /* Always points to base of allocation */
   char *curread, *curwrite; /* Iterators */
@@ -1097,7 +1256,8 @@ static void pollloop(const struct options *const o
 }
 #endif /* HAVE_POLL_H */
 
-static void loop(struct options *const o, struct progstate *const state)
+static void
+loop(struct options *const o, struct progstate *const state)
 {
   char *buf; /* Holds the whole buffer state->b*/
   char *curbuf; /* Iterates over buf when reading and leave after last
@@ -1429,19 +1589,22 @@ static void loop(struct options *const o, struct progstate *const state)
   free(palloc.allocated);
 }
 
-void cleanup(const struct options *const o, struct progstate *const state)
+void
+cleanup(const struct options *const o, struct progstate *const state)
 {
   if (o->p && state->pidfile_has_been_created)
     if (unlink(o->p) == -1)
       perror("Unlink pidfile failed - continuing");
 }
 
-static void print_version(void)
+static void
+print_version(void)
 {
   printf("%s\n", VERSION);
 }
 
-static void usage(void)
+static void
+usage(void)
 {
   fprintf(stderr, "cstream by Martin Cracauer - version " VERSION "\n");
   fprintf(stderr,
@@ -1488,11 +1651,15 @@ static void usage(void)
 	  "SIGUSR1 causes statistics to be written to stderr\n"
 	  "SIGUSR2 causes loop end after next buffer transfer\n"
 	  "<file>  if -i has not been used, specifies input file\n"
+          "-6 <n>  Use IPV6: -1 = don't, 1 = allow both, 2 = force v6\n"
+          "        On some platforms server mode 1 forces ipv6, as\n"
+          "        they don't open both v4 and v6 ports from one bind call.\n"
 	  );
   exit(1);
 }
 
-long long atoi_kmg(const char *const s)
+long long
+atoi_kmg(const char *const s)
 {
   long long res;
   char c;
@@ -1509,7 +1676,8 @@ long long atoi_kmg(const char *const s)
   return res;
 }
 
-int main(int argc, char *const argv[])
+int
+main(int argc, char *const argv[])
 {
   struct options o;
   struct progstate state;
@@ -1518,7 +1686,7 @@ int main(int argc, char *const argv[])
 
   default_options(&o);
 
-  while ((ch = getopt(argc, argv, "b:B:c:i:I:n:o:O:p:St:T:v:Vl")) != -1) {
+  while ((ch = getopt(argc, argv, "b:B:c:i:I:n:o:O:p:St:T:v:Vl6:")) != -1) {
     switch(ch) {
     case 'v': o.v = atoi(optarg); break;
     case 'b': blocksize = atoi_kmg(optarg); break;
@@ -1541,6 +1709,13 @@ int main(int argc, char *const argv[])
 #endif
       o.T = atoi(optarg);
       break;
+    case '6':
+      if (!atoi(optarg) != 0) {
+        fprintf(stderr, "-6 arg must be a number, -1, 1 or 2: '%s'\n", optarg);
+        exit(1);
+      }
+      o.six = atoi(optarg);
+      break;
     default: usage();
     }
   }
